structure RBTreeInsert :
sig
   include RBTREE
   val insert : int *
                ('color AColor, 'card ACard, ('condl, 'condn, 'condr) AGood) t ->
		(CBlack, CMany, CGood) t
end =
struct
   open RBTree

   fun restoreLeft (tree: ('color ABlack, 'card AMany, 
			   ('condl, 'condn, 'condr) ABrokenLeft) t) :
                   (CColor, CMany, CGood) t =
      caseBlackManyBrokenLeft tree
      {NodeBlack = fn (_,l,e,r) =>
      caseColorManyBrokenNode l 
      {Node = fn (c,ll,le,lr) => 
       let
	  fun ok (ll, lr) =
	     let
		val l' = nodeRedManyGood (Color.red, ll, le, lr)
		val tree = nodeBlackManyGood (Color.black, l', e, r)
	     in
		coerceColorManyGood tree
	     end
	  fun check_l lr =
	     caseColorCardGood ll
	     {Leaf = fn () => ok (leafBlackZeroGood, lr),
	      NodeRed = fn (_,lll,lle,llr) =>
                        let
			   val l' = nodeBlackManyGood (Color.black,lll,lle,llr)
			   val r' = nodeBlackManyGood (Color.black,lr,e,r)
			   val tree = nodeRedManyGood (Color.red,l',le,r')
			in
			   coerceColorManyGood tree
			end,
	      NodeBlack = fn x => ok (nodeBlackManyGood x, lr)}
	  fun check_r () =
	     caseColorCardGood lr
	     {Leaf = fn () => check_l leafBlackZeroGood,
	      NodeRed = fn (_,lrl,lre,lrr) =>
                        let
			   val l' = nodeBlackManyGood (Color.black, ll, le, lrl)
			   val r' = nodeBlackManyGood (Color.black, lrr, e, r)
			   val tree = nodeRedManyGood (Color.red, l', lre, r')
			in
			   coerceColorManyGood tree
			end,
	      NodeBlack = fn x => check_l (nodeBlackManyGood x)}
       in
	  Color.caseColor c
	  {Red = check_r,
	   Black = fn () => 
	           let
		      val l' = nodeBlackManyGood (Color.black, ll, le, lr)
		      val tree = nodeBlackManyGood (Color.black, l', e, r)
		   in
		      coerceColorManyGood tree
		   end}
       end}}

   fun restoreRight (tree: ('color ABlack, 'card AMany, 
			    ('condl, 'condn, 'condr) ABrokenRight) t) :
                    (CColor, CMany, CGood) t =
      caseBlackManyBrokenRight tree
      {NodeBlack = fn (_,l,e,r) =>
      caseColorManyBrokenNode r 
      {Node = fn (c,rl,re,rr) =>
       let
	  fun ok (rl, rr) =
	     let
		val r' = nodeRedManyGood (Color.red, rl, re, rr)
		val tree = nodeBlackManyGood (Color.black, l, e, r')
	     in
		coerceColorManyGood tree
	     end
	  fun check_l rr =
	     caseColorCardGood rl
	     {Leaf = fn () => ok (leafBlackZeroGood, rr),
	      NodeRed = fn (_,rll,rle,rlr) =>
                        let
			   val l' = nodeBlackManyGood (Color.black,l,e,rll)
			   val r' = nodeBlackManyGood (Color.black,rlr,re,rr)
			   val tree = nodeRedManyGood (Color.red,l',rle,r')
			in
			   coerceColorManyGood tree
			end,
	      NodeBlack = fn x => ok (nodeBlackManyGood x, rr)}
	  fun check_r () =
	     caseColorCardGood rr
	     {Leaf = fn () => check_l leafBlackZeroGood,
	      NodeRed = fn (_,rrl,rre,rrr) =>
                        let
			   val l' = nodeBlackManyGood (Color.black, l, e, rl)
			   val r' = nodeBlackManyGood (Color.black, rrl, rre, rrr)
			   val tree = nodeRedManyGood (Color.red, l', re, r')
			in
			   coerceColorManyGood tree
			end,
	      NodeBlack = fn x => check_l (nodeBlackManyGood x)}
       in
	  Color.caseColor c
	  {Red = check_r,
	   Black = fn () => 
	           let
		      val r' = nodeBlackManyGood (Color.black, rl, re, rr)
		      val tree = nodeBlackManyGood (Color.black, l, e, r')
		   in
		      coerceColorManyGood tree
		   end}
       end}}

   fun insert (elem: int,
	       tree: ('color AColor, 'card ACard, ('condl, 'condn, 'condr) AGood) t) :
              (CBlack, CMany, CGood) t =
      let
	fun insBlackCardGood (tree: (CBlack, CCard, CGood) t) : 
	                     (CColor, CMany, CGood) t =
	   caseBlackCardGood tree
	   {Leaf = fn () => (coerceColorManyGood o nodeRedManyGood)
	                    (Color.red, leafBlackZeroGood, elem, leafBlackZeroGood),
	    NodeBlack = fn (c,l,e,r) =>
                        if elem < e
			   then (restoreLeft o nodeBlackManyBrokenLeft)
			        (Color.black, insColorCardGood l, e, r)
			else if e < elem
			   then (restoreRight o nodeBlackManyBrokenRight)
			        (Color.black, l, e, insColorCardGood r)
			else (coerceColorManyGood o nodeBlackManyGood)
			     (Color.black, l, e, r)}
	and insBlackZeroGood (tree: (CBlack, CZero, CGood) t) :
	                     (CColor, CMany, CGood) t =
	   insBlackCardGood (coerceBlackCardGood tree)
	and insBlackManyGood (tree: (CBlack, CMany, CGood) t) :
	                     (CColor, CMany, CGood) t =
	   insBlackCardGood (coerceBlackCardGood tree)
	and insColorCardGood (tree: (CColor, CCard, CGood) t) :
                             (CColor, CMany, CBrokenNode) t =
	   caseColorCardGood tree
	   {Leaf = fn () => coerceColorManyBrokenNode
	                    (insBlackZeroGood leafBlackZeroGood),
	    NodeRed = fn (c,l,e,r) => 
	              if elem < e
			 then (coerceColorManyBrokenNode o nodeRedManyBrokenNode)
                              (Color.red, insBlackCardGood l, e, r)
		      else if e < elem
			 then (coerceColorManyBrokenNode o nodeRedManyBrokenNode)
                              (Color.red, l, e, insBlackCardGood r)
		      else (coerceColorManyBrokenNode o nodeRedManyGood)
			   (Color.red, l, e, r),
	    NodeBlack = fn (c,l,e,r) => 
                        coerceColorManyBrokenNode
	                (insBlackManyGood (nodeBlackManyGood (Color.black,l,e,r)))}
      in
	 caseColorManyBrokenNode 
	 (insColorCardGood (coerceColorCardGood tree))
	 {Node = fn (_,l,e,r) => nodeBlackManyGood (Color.black,l,e,r)}
      end
end
