signature  RBTREE =
sig
   structure Color : 
      sig
	 (* specialization type *)
	 type 'a t
	    
	 (* abstract types *)
	 type 'a AColor = {color: 'a} t
	 and 'a ARed   = {color: {red: 'a}} t
	 and 'a ABlack = {color: {black: 'a}} t
	    
	 (* concrete types *)
	 type CColor = unit AColor
	  and CRed   = unit ARed
	  and CBlack = unit ABlack

         (* constructors *)
         val red   : CRed
         val black : CBlack

         (* destructors *)
         val caseColor : 'a AColor -> {Red  : unit -> 'ret,
				       Black: unit -> 'ret} -> 'ret
         val caseRed   : 'a ARed   -> {Red  : unit -> 'ret} -> 'ret
         val caseBlack : 'a ABlack -> {Black: unit -> 'ret} -> 'ret

         (* coercions *)
         val coerceColor : 'a AColor -> CColor
      end
   type 'a AColor = 'a Color.AColor
    and 'a ARed   = 'a Color.ARed
    and 'a ABlack = 'a Color.ABlack
   type CColor = Color.CColor
    and CRed   = Color.CRed
    and CBlack = Color.CBlack
   structure Card :
      sig
	 (* specialization type *)
	 type 'a t

         (* abstract types *)
	 type 'a ACard = {card: 'a} t
	  and 'a AZero = {card: {zero: 'a}} t
	  and 'a AMany = {card: {many: 'a}} t

         (* concrete types *)
	 type CCard = unit ACard
	  and CZero = unit AZero
	  and CMany = unit AMany
      end
   type 'a ACard = 'a Card.ACard
    and 'a AZero = 'a Card.AZero
    and 'a AMany = 'a Card.AMany
   type CCard = Card.CCard
    and CZero = Card.CZero
    and CMany = Card.CMany
   structure Cond :
      sig
	 (* specialization type *)
	 type 'a t

	 (* abstract types *)
	 type ('a, 'b, 'c) ABroken = ('a * 'b * 'c) t
          and ('a, 'b, 'c) ABrokenLeft = ('a * {node: 'b} * {right: 'c}) t
          and ('a, 'b, 'c) ABrokenNode = ({left: 'a} * 'b * {right:'c}) t
          and ('a, 'b, 'c) ABrokenRight = ({left: 'a} * {right: 'b} * 'c) t
	  and ('a, 'b, 'c) AGood = ({left: 'a} * {node: 'b} * {right: 'c}) t

	 (* concrete type *)
	 type CBroken = (unit, unit, unit) ABroken
	  and CBrokenLeft = (unit, unit, unit) ABrokenLeft
	  and CBrokenNode = (unit, unit, unit) ABrokenNode
	  and CBrokenRight = (unit, unit, unit) ABrokenRight
	  and CGood = (unit, unit, unit) AGood
      end
   type ('a, 'b, 'c) ABroken = ('a, 'b, 'c) Cond.ABroken
    and ('a, 'b, 'c) ABrokenLeft = ('a, 'b, 'c) Cond.ABrokenLeft
    and ('a, 'b, 'c) ABrokenNode = ('a, 'b, 'c) Cond.ABrokenNode
    and ('a, 'b, 'c) ABrokenRight = ('a, 'b, 'c) Cond.ABrokenRight
    and ('a, 'b, 'c) AGood = ('a, 'b, 'c) Cond.AGood
   type CBroken = Cond.CBroken
    and CBrokenLeft = Cond.CBrokenLeft
    and CBrokenNode = Cond.CBrokenNode
    and CBrokenRight = Cond.CBrokenRight
    and CGood = Cond.CGood

   (*
    * datatype color_card_broken_tree 
    *          = Leaf 
    *          | Node of color * color_card_broken_tree * int * color_card_broken_tree 
    * withspec black_card_broken_tree 
    *          = Leaf
    *          | Node of black * color_card_broken_tree * int * color_card_broken_tree
    *      and black_zero_broken_tree
    *          = Leaf
    *      and color_many_broken_tree
    *          = Node of color * color_card_broken_tree * int * color_card_broken_tree
    *      and red_many_broken_tree
    *          = Node of red * color_card_broken_tree * int * color_card_broken_tree
    *      and black_many_broken_tree
    *          = Node of black * color_card_broken_tree * int * color_card_broken_tree
    *
    *      and color_many_brokennode_tree
    *          = Node of color * color_card_good_tree * int * color_card_good_tree
    *      and red_many_brokennode_tree
    *          = Node of red * color_card_good_tree * int * color_card_good_tree
    *
    *      and black_many_brokenleft_tree
    *          = Node of black * color_many_brokennode_tree * int * color_card_good_tree
    *
    *      and black_many_brokenright_tree
    *          = Node of black * color_card_good_tree * int * color_many_brokennode_tree
    *
    *      and color_card_good_tree 
    *          = Leaf
    *          | Node of red * black_card_good_tree * int * black_card_good_tree
    *          | Node of black * color_card_good_tree * int * color_card_good_tree
    *      and black_card_good_tree 
    *          = Leaf
    *          | Node of black * color_card_good_tree * int * color_card_good_tree
    *      and black_zero_good_tree 
    *          = Leaf
    *      and color_many_good_tree
    *          = Node of red * black_card_good_tree * int * black_card_good_tree
    *          | Node of black * color_card_good_tree * int * color_card_good_tree
    *      and red_many_good_tree 
    *          = Node of red * black_card_good_tree * int * black_card_good_tree
    *      and black_many_good_tree
    *          = Node of black * color_card_good_tree * int * color_card_good_tree
    *)

   (* specialization type *)
   type ('color, 'card, 'cond) t

   (* constructors *)
   val nodeRedManyBrokenNode :
      'color ARed * 
      ('colorL AColor, 'cardL ACard, ('condlL, 'condnL, 'condrL) AGood) t *
      int *
      ('colorR AColor, 'cardR ACard, ('condlR, 'condnR, 'condrR) AGood) t ->
      (CRed, CMany, CBrokenNode) t
      
   val nodeBlackManyBrokenLeft :
      'color ABlack *
      ('colorL AColor, 'cardL AMany, ('condlL, 'condnL, 'condrL) ABrokenNode) t *
      int *
      ('colorR AColor, 'cardR ACard, ('condlR, 'condnR, 'condrR) AGood) t ->
      (CBlack, CMany, CBrokenLeft) t
   val nodeBlackManyBrokenRight :
      'color ABlack *
      ('colorL AColor, 'cardL ACard, ('condlL, 'condnL, 'condrL) AGood) t *
      int *
      ('colorR AColor, 'cardR AMany, ('condlR, 'condnR, 'condrR) ABrokenNode) t ->
      (CBlack, CMany, CBrokenRight) t

   val leafBlackZeroGood :
      (CBlack, CZero, CGood) t
   val nodeRedManyGood :
      'color ARed * 
      ('colorL ABlack, 'cardL ACard, ('condlL, 'condnL, 'condrL) AGood) t *
      int *
      ('colorR ABlack, 'cardR ACard, ('condlR, 'condnR, 'condrR) AGood) t ->
      (CRed, CMany, CGood) t
   val nodeBlackManyGood : 
      'color ABlack * 
      ('colorL AColor, 'cardL ACard, ('condlL, 'condnL, 'condrL) AGood) t *
      int *
      ('colorR AColor, 'cardR ACard, ('condlR, 'condnR, 'condrR) AGood) t ->
      (CBlack, CMany, CGood) t

   (* destructors *)
   val caseColorCardBroken :
      ('color AColor, 'card ACard, ('condl, 'condn, 'condr) ABroken) t ->
      {Leaf: unit -> 'ret,
       Node: CColor *
             (CColor, CCard, CBroken) t *
	     int *
	     (CColor, CCard, CBroken) t -> 'ret} -> 'ret

   val caseColorManyBrokenNode :
      ('color AColor, 'card AMany, ('condl, 'condn, 'condr) ABrokenNode) t ->
      {Node: CColor * 
             (CColor, CCard, CGood) t *
	     int *
	     (CColor, CCard, CGood) t -> 'ret} -> 'ret

   val caseBlackManyBrokenLeft :
      ('color ABlack, 'card AMany, ('condl, 'condn, 'condr) ABrokenLeft) t ->
      {NodeBlack: CBlack * 
                  (CColor, CMany, CBrokenNode) t *
		  int *
		  (CColor, CCard, CGood) t -> 'ret} -> 'ret
   val caseBlackManyBrokenRight :
      ('color ABlack, 'card AMany, ('condl, 'condn, 'condr) ABrokenRight) t ->
      {NodeBlack: CBlack * 
                  (CColor, CCard, CGood) t *
		  int *
		  (CColor, CMany, CBrokenNode) t -> 'ret} -> 'ret

   val caseColorCardGood :
      ('color AColor, 'card ACard, ('condl, 'condn, 'condr) AGood) t ->
      {Leaf: unit -> 'ret,
       NodeRed: CRed * 
                (CBlack, CCard, CGood) t *
		int *
		(CBlack, CCard, CGood) t -> 'ret,
       NodeBlack: CBlack * 
                  (CColor, CCard, CGood) t *
		  int *
		  (CColor, CCard, CGood) t -> 'ret} -> 'ret
   val caseColorManyGood :
      ('color AColor, 'card AMany, ('condl, 'condn, 'condr) AGood) t ->
      {NodeRed: CRed * 
                (CBlack, CCard, CGood) t *
		int *
		(CBlack, CCard, CGood) t -> 'ret,
       NodeBlack: CBlack * 
                  (CColor, CCard, CGood) t *
		  int *
		  (CColor, CCard, CGood) t -> 'ret} -> 'ret
   val caseBlackCardGood :
      ('color ABlack, 'card ACard, ('condl, 'condn, 'condr) AGood) t ->
      {Leaf: unit -> 'ret,
       NodeBlack: CBlack * 
                  (CColor, CCard, CGood) t *
		  int *
		  (CColor, CCard, CGood) t -> 'ret} -> 'ret

   (* coercions *)
   val coerceColorCardBroken :
      ('color AColor, 'card ACard, ('condl, 'condn, 'condr) ABroken) t ->
      (CColor, CCard, CBroken) t
   val coerceColorManyBrokenNode :
      ('color AColor, 'card AMany, ('condl, 'condn, 'condr) ABrokenNode) t ->
      (CColor, CMany, CBrokenNode) t
   val coerceColorCardGood :
      ('color AColor, 'card ACard, ('condl, 'condn, 'condr) AGood) t ->
      (CColor, CCard, CGood) t
   val coerceBlackCardGood :
      ('color ABlack, 'card ACard, ('condl, 'condn, 'condr) AGood) t ->
      (CBlack, CCard, CGood) t
   val coerceColorManyGood :
      ('color AColor, 'card AMany, ('condl, 'condn, 'condr) AGood) t ->
      (CColor, CMany, CGood) t
end
