structure RBTree :> RBTREE =
struct
   datatype color = Red | Black
   datatype tree = Leaf | Node of color * tree * int * tree
   fun fail _ = raise Match

   structure Color =
      struct
	 type 'a t = color

	 (* abstract types *)
	 type 'a AColor = {color: 'a} t
	 and 'a ARed   = {color: {red: 'a}} t
	 and 'a ABlack = {color: {black: 'a}} t
	    
	 (* concrete types *)
	 type CColor = unit AColor
	  and CRed   = unit ARed
	  and CBlack = unit ABlack

         (* constructors *)
         val red   = Red
         val black = Black
	    
         (* destructors *)
	 fun caseColor c {Red=r,Black=b} =
	    case c of Red => r () | Black => b ()
	 fun caseRed c {Red=r} = caseColor c {Red=r,Black=fail}
	 fun caseBlack c {Black=b} = caseColor c {Red=fail,Black=b}
	    
         (* coercions *)
	 fun coerce c = c
	 fun coerceColor c = c
      end
   type 'a AColor = 'a Color.AColor
    and 'a ARed   = 'a Color.ARed
    and 'a ABlack = 'a Color.ABlack
   type CColor = Color.CColor
    and CRed   = Color.CRed
    and CBlack = Color.CBlack
   structure Card =
      struct
	 (* specialization type *)
	 type 'a t = unit

         (* abstract types *)
	 type 'a ACard = {card: 'a} t
	  and 'a AZero = {card: {zero: 'a}} t
	  and 'a AMany = {card: {many: 'a}} t

         (* concrete types *)
	 type CCard = unit ACard
	  and CZero = unit AZero
	  and CMany = unit AMany
      end
   type 'a ACard = 'a Card.ACard
    and 'a AZero = 'a Card.AZero
    and 'a AMany = 'a Card.AMany
   type CCard = Card.CCard
    and CZero = Card.CZero
    and CMany = Card.CMany
   structure Cond =
      struct
	 (* specialization type *)
	 type 'a t = unit

	 (* abstract types *)
	 type ('a, 'b, 'c) ABroken = ('a * 'b * 'c) t
          and ('a, 'b, 'c) ABrokenLeft = ('a * {node: 'b} * {right: 'c}) t
          and ('a, 'b, 'c) ABrokenNode = ({left: 'a} * 'b * {right:'c}) t
          and ('a, 'b, 'c) ABrokenRight = ({left: 'a} * {right: 'b} * 'c) t
	  and ('a, 'b, 'c) AGood = ({left: 'a} * {node: 'b} * {right: 'c}) t

	 (* concrete type *)
	 type CBroken = (unit, unit, unit) ABroken
	  and CBrokenLeft = (unit, unit, unit) ABrokenLeft
	  and CBrokenNode = (unit, unit, unit) ABrokenNode
	  and CBrokenRight = (unit, unit, unit) ABrokenRight
	  and CGood = (unit, unit, unit) AGood
      end
   type ('a, 'b, 'c) ABroken = ('a, 'b, 'c) Cond.ABroken
    and ('a, 'b, 'c) ABrokenLeft = ('a, 'b, 'c) Cond.ABrokenLeft
    and ('a, 'b, 'c) ABrokenNode = ('a, 'b, 'c) Cond.ABrokenNode
    and ('a, 'b, 'c) ABrokenRight = ('a, 'b, 'c) Cond.ABrokenRight
    and ('a, 'b, 'c) AGood = ('a, 'b, 'c) Cond.AGood
   type CBroken = Cond.CBroken
    and CBrokenLeft = Cond.CBrokenLeft
    and CBrokenNode = Cond.CBrokenNode
    and CBrokenRight = Cond.CBrokenRight
    and CGood = Cond.CGood

   type ('color,'card,'cond) t = tree
       
   (* constructors *)
   val nodeRedManyBrokenNode = Node

   val nodeBlackManyBrokenLeft = Node
   val nodeBlackManyBrokenRight = Node

   val leafBlackZeroGood = Leaf
   val nodeRedManyGood = Node
   val nodeBlackManyGood = Node

   (* destructors *)
   fun caseTree t {Leaf=l,Node=n} =
      case t of Leaf => l () | Node x => n x
   fun caseColorCardBroken t {Leaf=l,Node=n} = caseTree t {Leaf=l,Node=n}
   fun caseColorManyBrokenNode t {Node=n} = caseTree t {Leaf=fail,Node=n}
   fun caseBlackManyBrokenLeft t {NodeBlack=nb} = caseTree t {Leaf=fail,Node=nb}
   fun caseBlackManyBrokenRight t {NodeBlack=nb} = caseTree t {Leaf=fail,Node=nb}
   fun caseColorCardGood t {Leaf=l,NodeRed=nr,NodeBlack=nb} = 
      caseTree t {Leaf=l,Node=fn node as (c,_,_,_) => 
		  case c of Red => nr node | Black => nb node}
   fun caseColorManyGood t {NodeRed=nr,NodeBlack=nb} =
      caseTree t {Leaf=fail,Node=fn node as (c,_,_,_) => 
		  case c of Red => nr node | Black => nb node}
   fun caseBlackCardGood t {Leaf=l,NodeBlack=nb} = caseTree t {Leaf=l,Node=nb}

   (* coercions *) 
   fun coerce t = t
   val coerceColorCardBroken = coerce
   val coerceColorManyBrokenNode = coerce
   val coerceColorCardGood = coerce
   val coerceColorManyGood = coerce
   val coerceBlackCardGood = coerce
end
